unit uCombustible;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

uses
  SysUtils,
  ucosa,
  ucosaparticipedemercado,
  uFichasLPD, xMatDefs, ufechas, uconstantesSimSEE, uglobs;

type

  TFichaCombustible = class(TFichaLPD)
  public

    precio: NReal;
    PCI: NReal;
    PCS: NReal;

    constructor Create(capa: integer; fecha: TFecha; periodicidad: TPeriodicidad;
      precio, PCI, PCS: NReal);
    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
    procedure generarLineaResumen(var archi: TextFile); override;
    function InfoAd_: string; override;
    procedure Free; override;
  end;

  TClaseDeCombustible = class of TCombustible;

  TCombustible = class(TCosaParticipeDeMercado)
  public
    pa: TFichaCombustible;
    globs: TGlobs;

    densidad: NReal;
    unidad: string;

    constructor Create(capa: integer; nombre: string; lpd: TFichasLPD;
      densidad: NReal; unidad: string); reintroduce;
    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
    function InfoAd_: string; override;
    procedure RegistrarParametrosDinamicos;
    class function DescClase: string; override;
  end;

procedure AlInicio;
procedure AlFinal;


resourcestring
  rsCombusitbles = 'Combustibles';


implementation

//------------------------------
// Métodos de TFichaCombustible
//==============================
constructor TFichaCombustible.Create(capa: integer; fecha: TFecha; periodicidad: TPeriodicidad;
  precio, PCI, PCS: NReal);
begin
  inherited Create(capa, fecha, periodicidad);
  self.precio := precio;
  self.PCI := PCI;
  self.PCS := PCS;
end;

constructor TFichaCombustible.Create_ReadFromText(f: TArchiTexto);
begin
  inherited Create_ReadFromText(f);
  f.IniciarLecturaRetrasada;

  f.rd('precio', precio);
  f.rd('PCI', PCI);
  f.rd('PCS', PCS);

  f.EjecutarLectura;
end;

procedure TFichaCombustible.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wr('precio', precio, uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);
  f.wr('PCI', PCI, uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);
  f.wr('PCS', PCS, uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);
end;

procedure TFichaCombustible.generarLineaResumen(var archi: TextFile);
begin
  Write(archi, FloatToStrF(precio, formatoReales, 8, 1), #9,
    FloatToStrF(PCI, formatoReales, 8, 1), #9,
    FloatToStrF(PCS, formatoReales, 8, 2), #9);
end;

function TFichaCombustible.InfoAd_: string;
begin
  Result := 'precio= ' + FloatToStrF(precio, ffGeneral, 10, 1) +
    ' USD, ' + 'PCI= ' + FloatToStrF(PCI, ffGeneral, 10, 1) + ' MW, ' +
    'PCS= ' + FloatToStrF(PCS, ffGeneral, 10, 1) + ' MW, ';
end;

procedure TFichaCombustible.Free;
begin
  inherited Free;
end;

//------------------------------
// Métodos de TCombustible
//==============================

constructor TCombustible.Create(capa: integer; nombre: string; lpd: TFichasLPD;
  densidad: NReal; unidad: string);
begin
  inherited Create(capa, nombre);
  self.lpd := lpd;
  self.densidad := densidad;
  self.unidad := unidad;
end;

constructor TCombustible.Create_ReadFromText(f: TArchiTexto);
begin
  inherited Create_ReadFromText(f);
  f.IniciarLecturaRetrasada;

  f.rd('densidad', densidad);
  f.rd('unidad', unidad);
  f.rd('lpd', TCosa(lpd));

  f.EjecutarLectura;
end;

procedure TCombustible.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wr('densidad', densidad, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr('unidad', unidad);
  f.wr('lpd', TCosa(lpd));
end;

function TCombustible.InfoAd_: string;
begin
  Result := '';
end;

procedure TCombustible.registrarParametrosDinamicos;
begin
  lpd.expandirFichas(globs);
  lpd.RegistrarFichasAActualizar(Self, globs.ActualizadorLPD, @pA, nil);
end;

class function TCombustible.DescClase: string;
begin
  Result := rsCombusitbles;
end;

procedure AlInicio;
begin
  registrarClaseDeCosa(TCombustible.ClassName, TCombustible);
  registrarClaseDeCosa(TFichaCombustible.ClassName, TFichaCombustible);
end;

procedure AlFinal;
begin
end;

end.
