unit uDataSetGenerico;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface
uses
  Classes,
  sysutils;

type
  TDescripcionDeCampo = record
    nombre, tipo: string;
  end;
  TDescripcionDeCampos = array of TDescripcionDeCampo;

  TDataSetGenerico = class;

  TDataRecord = class
    private
    public
      padre: TDataSetGenerico;
      valores: array of String;
      freePadre: boolean;

      constructor Create( padre: TDataSetGenerico );
      procedure SetValById( id: integer; val: string );
      function GetByIdAsString( id: integer ): string;
      function GetByNameAsString( nombre: string ): string;
      function GetByNameAsFloat( nombre: string ): double;
      function GetByNameAsInt( nombre: string ): integer;
      procedure Free;

      property valor[i: Integer]: String read GetByIdAsString write SetValById; default;
  end;

  TDataRecords = array of TDataRecord;

  TDataSetGenerico = class
    private
      kFichaActiva: integer;
    public
      nrows, nfields: Integer;
      descripcionDeCampos: TDescripcionDeCampos;
      resultados: TDataRecords;

      Constructor Create(nrows, nfields: Integer);

      function first: TDataRecord;
      function next: TDataRecord;
      function go(kRec: integer ): TDataRecord;

      function FieldName( kField: integer ): string;
      function FieldType( kField: integer ): string;
      function kOfField( nombre: string ): integer; // -1 si no existe
      function eof: boolean;
      procedure Free;

      function getFila(i: Integer): TDataRecord;
      procedure setFila(i: Integer; fila: TDataRecord);

      property filas[i: Integer]: TDataRecord read getFila write setFila; default;
  end;

  TCeldas = array of array of String;
  TDataSetStringGrid = class(TDataSetGenerico)
    public
      //Si nCampos es -1 se toma el largo de encabezados como nFields
      Constructor Create(encabezados: array of String; nFilas: Integer; nCampos: Integer = -1); reintroduce; overload;
      //
      Constructor Create(encabezados: array of String; celdas: TCeldas); reintroduce; overload;
  end;

implementation

//----------------------
//Mtodos de TDataRecord
//----------------------

constructor TDataRecord.Create( padre: TDataSetGenerico );
begin
  inherited Create();
  self.padre:= padre;
  self.freePadre:= false; // por defecto las fichas no liberan al papa
  SetLength(valores, padre.nfields);
end;

procedure TDataRecord.SetValById( id: integer; val: string );
begin
  valores[id]:= val;
end;

function TDataRecord.GetByIdAsString( id: integer ): string;
begin
  result:= valores[id];
end;

function TDataRecord.GetByNameAsString( nombre: string ): string;
var
  k: integer;
begin
  k:= padre.kOfField(nombre);
  if k < 0 then
     raise Exception.Create('Error, GetByNameAsString, campo no conocido: '+nombre );
  result:= valores[k];
end;

function TDataRecord.GetByNameAsFloat( nombre: string ): double;
var
  k: integer;
  res: double;
  rescode: integer;
begin
  k:= padre.kOfField(nombre);
  if k < 0 then
     raise Exception.Create('Error, GetByNameAsFloat, campo no conocido: '+nombre );
  val( valores[k], res, rescode );
  if rescode <> 0 then
    result:= 0
  else
    result:= res;
end;

function TDataRecord.GetByNameAsInt( nombre: string ): integer;
var
  k: integer;
  res: integer;
  rescode: integer;
begin
  k:= padre.kOfField(nombre);
  if k < 0 then
     raise Exception.Create('Error, GetByNameAsInt, campo no conocido: '+nombre );
  val( valores[k], res, rescode );
  if rescode <> 0 then
    result:= 0
  else
    result:= res;
end;


procedure TDataRecord.Free;
begin
  if freePadre  then
    Padre.Free;

  SetLength(valores, 0);
  inherited Free;
end;

//---------------------------
//Mtodos de TDataSetGenerico
//---------------------------

Constructor TDataSetGenerico.Create(nrows, nfields: Integer);
begin
  inherited Create;
  self.nrows:= nrows;
  self.nfields:= nfields;

  SetLength(descripcionDeCampos, nfields);
  SetLength(resultados, nrows);
  kFichaActiva:= -1;
end;

function TDataSetGenerico.first: TDataRecord;
begin
  kFichaActiva:= 0;
  if nrows > 0 then
    result:= resultados[0]
  else
    result:= nil;
end;

function TDataSetGenerico.next: TDataRecord;
begin
  kFichaActiva:= kFichaActiva + 1;
  if kFichaActiva < length(resultados) then
    result:= resultados[kFichaActiva]
  else
  begin
    kFichaActiva:= length( resultados );
    result:= NIL;
  end;
end;

function TDataSetGenerico.go(kRec: integer ): TDataRecord;
begin
  if ( kRec >= 0 ) and ( kRec < length( resultados ) ) then
  begin
    kFichaActiva:= kRec;
    result:= resultados[kRec];
  end
  else
    result:= nil;
end;

function TDataSetGenerico.FieldName( kField: integer ): string;
begin
  result:= self.descripcionDeCampos[kField].nombre;
end;

function TDataSetGenerico.FieldType( kField: integer ): string;
begin
  result:= self.descripcionDeCampos[kField].tipo;
end;

function TDataSetGenerico.kOfField( nombre: string): integer; // -1 si no existe
var
  i, res: Integer;
begin
  res:= -1;
  for i:= 0 to high(descripcionDeCampos) do
    if descripcionDeCampos[i].nombre = nombre then
    begin
      res:= i;
      break;
    end;
  result:= res;
end;

function TDataSetGenerico.eof: boolean;
begin
  result:= (kFichaActiva >= Length(resultados)) or
           (kFichaActiva < 0);
end;

procedure TDataSetGenerico.Free;
var
  i: Integer;
  f: TDataRecord;
begin
  SetLength(descripcionDeCampos, 0);
  for i:= 0 to high(resultados) do
  begin
    f:= resultados[i];
    if not f.freePadre then
      f.Free;
  end;
  SetLength(resultados, 0);
  inherited Free;
end;

function TDataSetGenerico.getFila(i: Integer): TDataRecord;
begin
  result:= Resultados[i];
end;

procedure TDataSetGenerico.setFila(i: Integer; fila: TDataRecord);
begin
  resultados[i]:= fila;
end;


//-----------------------------
//Mtodos de TDataSetStringGrid
//-----------------------------

Constructor TDataSetStringGrid.Create(encabezados: array of string; nFilas, nCampos: Integer);
var
  i: Integer;
begin
  if nCampos < 0 then
    inherited Create(nFilas, Length(encabezados))
  else
    inherited Create(nFilas, nCampos);
    
  for i:= 0 to high(encabezados) do
  begin
    descripcionDeCampos[i].nombre:= encabezados[i];
    descripcionDeCampos[i].tipo:= 'string';
  end;
  for i:= length(encabezados) to nfields - 1 do
  begin
    descripcionDeCampos[i].nombre:= '';
    descripcionDeCampos[i].tipo:= 'string';
  end;

  for i:= 0 to nFilas - 1 do
    resultados[i]:= TDataRecord.Create(self);
end;

Constructor TDataSetStringGrid.Create(encabezados: array of string; celdas: TCeldas);
var
  i, j: Integer;
begin
  inherited Create(Length(celdas), length(encabezados));
  for i:= 0 to nfields - 1 do
  begin
    descripcionDeCampos[i].nombre:= encabezados[i];
    descripcionDeCampos[i].tipo:= 'string';
  end;

  for i:= 0 to nrows - 1 do
  begin
    resultados[i]:= TDataRecord.Create(self);
    for j:= 0 to nfields - 1 do
      resultados[i].SetValById(j, celdas[i][j]);
  end;
end;

end.
